
#include <propeller2.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#define _DIAG

enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,

	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	SD_BASE_PIN = 16,

	PIN_DAT    = SD_BASE_PIN | 3<<6,
	PIN_DAT0   = SD_BASE_PIN+0,
	PIN_DAT1   = SD_BASE_PIN+1,
	PIN_DAT2   = SD_BASE_PIN+2,
	PIN_DAT3   = SD_BASE_PIN+3,
	PIN_CMD    = SD_BASE_PIN+4,
	PIN_CLK    = SD_BASE_PIN+5,
	PIN_LED    = SD_BASE_PIN+6,  // command input smartpin, also outputs to Activity LED
	PIN_CD     = SD_BASE_PIN+7,  // Card-Detect switch input, doubles up as PIN_CMD start-bit detect

	CLK_DIV = 2,
};


static uint32_t  rca;
static uint8_t  resp[20];
static uint8_t  buff[2000];

/*
    SD Responses
   ==============
    R1   48 bits    Status (32-bit)
    R1b  48 bits, and busy on DAT0
    R2  136 bits    CID or CSD
    R3   48 bits    OCR (32-bit)
    R6   48 bits    RCA (16-bit) and Status (16-bit)
    R7   48 bits    volts (4-bit)
*/

#ifdef _DIAG
static uint32_t  crc7( uint8_t * buf, size_t len )    // SD spec 4.5
{
    uint32_t  val, crc = 0;

    __asm {
// Reference code courtesy of Ariba
cr7lp
		rdword	val, buf
		add	buf, #2
		movbyts	val, #0b00_01_10_11
		setq	val
		crcnib	crc, #0x48    // CCITT polynomial is 1 + x3 + x7 (0x09 reversed for CRCNIB)
		crcnib	crc, #0x48
		sub	len, #1   wcz
	if_ne	crcnib	crc, #0x48
	if_ne	crcnib	crc, #0x48
	if_ne	djnz	len, #cr7lp

		rev	crc         // correct the bit order to match standard
		shr	crc, #24
		or	crc, #1
    }
    return crc;
}
#endif


static uint32_t rxblock( uint8_t * buf )
{
    uint32_t  i, j;

    _pinstart( PIN_CLK, P_PWM_SMPS | P_OE | P_INVERT_OUTPUT |    // clock gen stops when PIN_DAT3 goes low
                (PIN_DAT3 - PIN_CLK & 7)<<28 | P_INVERT_A |    // smartA input select
                (PIN_DAT3 - PIN_CLK & 7)<<24 | P_INVERT_B,    // smartB input select
                1 | 8<<16, 4 );    // clock rate of sysclock/8
    _pinf( PIN_LED );    // reset timeout counter
    _pinl( PIN_LED );

    __asm volatile {    // "volatile" enforces use of FCACHE
		setse1	#0b100<<6 | PIN_DAT3    // trigger on low level - start bit
		wrfast	xfrq, buf
		setxfrq	xfrq        // set sysclock/1 for lead-in timing
		waitse1        // wait for start bit

		dirl	#PIN_CLK
		wrpin	mclk, #PIN_CLK    // revert to regular clock gen
		wxpin	mclkr, #PIN_CLK

		dirh	#PIN_CLK    // clock period timing starts here
		wypin	clocks, #PIN_CLK    // first clock pulse outputs during second period
		xinit	mleadin, #0    // lead-in delay from here at sysclock/1
		setq	mnco        // streamer transfer rate (takes effect with buffered command below)
		xcont	mca, #0     // rx buffered-op aligned with clock via lead-in

		// twiddle thumbs
//		rdpin	j, #PIN_LED

		waitxfi    // wait for rx data completion before resuming hubexec
		ret

mclk		long	P_PULSE | P_OE | P_INVERT_OUTPUT
mclkr		long	CLK_DIV | CLK_DIV/2<<16
clocks		long	1056
xfrq		long	0x8000_0000
mnco		long	0x8000_0000UL / CLK_DIV   // + (0x8000_0000UL % CLK_DIV > 0UL ? 1 : 0)    // round up upon non-zero remainder
mleadin		long	X_IMM_32X1_1DAC1 | CLK_DIV * 2 + 1    // no clock occurs for first clock period
mca		long	X_4P_1DAC4_WFBYTE | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 1056
    }

#ifdef _DIAG
    printf( "   CRC = %02x\n", crc7(resp,5) );
//    printf( "   CRC = %02x  Data latency = %d\n", crc7(resp,5), j );

//    for( i = 0; i <= 799/32; i++ )  {    // diag
        for( j = 0; j < 32; j++ )
//            printf( " %02x", buf[i*32+j] );
            printf( " %02x", buf[j] );
        puts("        ");
//        puts("                                        ");
//    }
#endif

    return 1;
}

/*
static uint32_t rxcmd( size_t len )
{
    size_t  i;

    _pinf( PIN_CMD );
    _wrpin( PIN_CMD, 0 );
    _pinstart( PIN_CLK, P_PWM_SMPS | P_OE | P_INVERT_OUTPUT |    // clock gen stops when PIN_DAT3 goes low
                (PIN_CMD - PIN_CLK & 7)<<28 | P_INVERT_A |    // smartA input select
                (PIN_CMD - PIN_CLK & 7)<<24 | P_INVERT_B,    // smartB input select
                1 | 8<<16, 4 );    // clock rate of sysclock/8

    __asm volatile {    // "volatile" enforces use of FCACHE
		setse1	#0b100<<6 | PIN_CMD    // trigger on low level - start bit
		getct	i
		loc	pb, #resp    // response buffer address in hubRAM, static location
		wrfast	xfrq, pb
		setxfrq	xfrq        // set sysclock/1 for lead-in timing
		mov	pb, len
		shl	pb, #3    // number of bits in the response
		setword	mca, pb, #0    // fill out 16-bit I/O cycle count of streamer
		add	pb, #8    // 8 extra clock pulses
		add	i, ##64*100
		setq	i
		waitse1   wc      // wait for start bit, C set if timed-out 

		dirl	#PIN_CLK
		wrpin	mclk, #PIN_CLK    // revert to regular clock gen
		wxpin	mclkr, #PIN_CLK

		dirh	#PIN_CLK    // clock period timing starts here
	if_nc	wypin	pb, #PIN_CLK    // first clock pulse outputs during second period
	if_nc	xinit	mleadin, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	mnco        // streamer transfer rate (takes effect with buffered command below)
	if_nc	xcont	mca, #0     // rx buffered-op aligned with clock via lead-in

		// twiddle thumbs
	if_nc	mov	i, #0

	if_nc	waitxfi    // wait for rx data completion before resuming hubexec
		ret

mclk		long	P_PULSE | P_OE | P_INVERT_OUTPUT
mclkr		long	CLK_DIV | CLK_DIV/2<<16
xfrq		long	0x8000_0000
mnco		long	0x8000_0000UL / CLK_DIV   // + (0x8000_0000UL % CLK_DIV > 0UL ? 1 : 0)    // round up upon non-zero remainder
mleadin		long	X_IMM_32X1_1DAC1 | CLK_DIV    // sample first non-clock period to collect the start-bit
mca		long	X_1P_1DAC1_WFBYTE | PIN_CMD<<17 | X_PINS_ON | X_ALT_ON
    }

    printf( " i=%x", i );
    if( i )
        return 0;    // timed-out

#ifdef _DIAG
    for( i = 0; i < len; i++ )
        printf( " %02x", resp[i] );
#endif

    return 1;    // card has responded
}
*/
/*
static uint32_t rxcmd( size_t len )
{
    size_t  i;

    _pinf( PIN_CMD );
    _wrpin( PIN_CMD, 0 );

    __asm volatile {    // "volatile" enforces use of FCACHE
		loc	pb, #resp    // response buffer address in hubRAM, static location
		wrfast	xfrq, pb
		setxfrq	xfrq        // set sysclock/1 for lead-in timing

		mov	pb, #64
.loop1
		wypin	#1, #PIN_CLK    // first clock pulse outputs during second period
		waitx	#CLK_DIV * 2 + 8
		testp	#PIN_CMD   wc
	if_c	djnz	pb, #.loop1    // loop while CMD pin is high

		mov	pb, len
		shl	pb, #3    // number of bits in response
		setword	mca, pb, #0    // fill out 16-bit I/O cycle count of streamer
		add	pb, #8    // 8 extra clock pulses

		dirl	#PIN_CLK
		dirh	#PIN_CLK    // clock period timing starts here
	if_nc	wypin	pb, #PIN_CLK    // first clock pulse outputs during second period
	if_nc	xinit	mleadin, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	mnco        // streamer transfer rate (takes effect with buffered command below)
	if_nc	xcont	mca, #0     // rx buffered-op aligned with clock via lead-in

		// twiddle thumbs
	if_nc	mov	i, #0

	if_nc	waitxfi    // wait for rx data completion before resuming hubexec
		ret

xfrq		long	0x8000_0000
mnco		long	0x8000_0000UL / CLK_DIV // + (0x8000_0000UL % CLK_DIV > 0UL ? 1 : 0)    // round up upon non-zero remainder
mleadin		long	X_IMM_32X1_1DAC1 | CLK_DIV    // sample first non-clock period to collect start-bit
mca		long	X_1P_1DAC1_WFBYTE | PIN_CMD<<17 | X_PINS_ON | X_ALT_ON
    }

    printf( " i=%x", i );
    if( i )
        return 0;    // timed-out

#ifdef _DIAG
    for( i = 0; i < len; i++ )
        printf( " %02x", resp[i] );
#endif

    return 1;    // card has responded
}
*/
/*
static uint32_t rxcmd4( size_t len )
{
    size_t  i;

    _pinf( PIN_CMD );
    _wrpin( PIN_CMD, 0 );
    _pinstart( PIN_CLK, P_PWM_SMPS | P_OE | P_INVERT_OUTPUT |    // SD clock gen smartpin
                (PIN_CMD - PIN_CLK & 7)<<28 | P_INVERT_A |    // smartA input select
                (PIN_CMD - PIN_CLK & 7)<<24 | P_INVERT_B,    // smartB input select
                1 | 8<<16, 4 );
    _pinf( PIN_LED );    // reset timeout counter
    _pinl( PIN_LED );

    while( _pinr( PIN_CMD ) && (_rdpin( PIN_LED ) < 64) );    // wait for CMD low - found start bit

    _pinstart( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT,    // reconfig back to regular clock gen
                CLK_DIV | (CLK_DIV / 2)<<16, 0 );

    __asm {
.loop0
		rdpin	i, #PIN_LED
		cmp	i, #64   wc    // if clocks < 64
		testp	#PIN_CMD   wz    // if CMD
if_c_or_z	jmp	#.loop0    // loop while clocks < 64 and CMD pin high

		setse1	#001<<6 | PIN_CMD
		loc	ptrb, #resp

		waitse1
		rdpin	pb, #PIN_CMD
		wxpin	#15, #PIN_CMD    // shifter to 16 bits for remainder of response
		rev	pb
		zerox	pb, #14    // add the start bit (bit15 = 0)
		movbyts	pb, #0b00_01
		wrword	pb, ptrb++

		mov	i, len
		sub	i, #1
		shr	i, #1
.loop1
		waitse1
		rdpin	pb, #PIN_CMD
		rev	pb
		movbyts	pb, #0b00_01
		wrword	pb, ptrb++
		djnz	i, #.loop1
    }
    __asm volatile {    // "volatile" enforces use of FCACHE
		loc	pb, #resp    // response buffer address in hubRAM, static location
		wrfast	xfrq, pb
		setxfrq	xfrq        // set sysclock/1 for lead-in timing

		mov	pb, #64
.loop1
		wypin	#1, #PIN_CLK    // first clock pulse outputs during second period
		waitx	#CLK_DIV * 2 + 8
		testp	#PIN_CMD   wc
	if_c	djnz	pb, #.loop1    // loop while CMD pin is high

		mov	pb, len
		shl	pb, #3    // number of bits in the response
		setword	mca, pb, #0    // fill out 16-bit I/O cycle count of streamer
		add	pb, #8    // 8 extra clock pulses

		dirl	#PIN_CLK
		dirh	#PIN_CLK    // clock period timing starts here
	if_nc	wypin	pb, #PIN_CLK    // first clock pulse outputs during second period
	if_nc	xinit	mleadin, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	mnco        // streamer transfer rate (takes effect with buffered command below)
	if_nc	xcont	mca, #0     // rx buffered-op aligned with clock via lead-in

		// twiddle thumbs
	if_nc	mov	i, #0

	if_nc	waitxfi    // wait for rx data completion before resuming hubexec
		ret

xfrq		long	0x8000_0000
mnco		long	0x8000_0000UL / CLK_DIV // + (0x8000_0000UL % CLK_DIV > 0UL ? 1 : 0)    // round up upon non-zero remainder
mleadin		long	X_IMM_32X1_1DAC1 | CLK_DIV    // sample first non-clock period to collect start-bit
mca		long	X_1P_1DAC1_WFBYTE | PIN_CMD<<17 | X_PINS_ON | X_ALT_ON
    }

    printf( " i=%x", i );
    if( i )
        return 0;    // timed-out

#ifdef _DIAG
    for( i = 0; i < len; i++ )
        printf( " %02x", resp[i] );
#endif

    return 1;    // card has responded
}
*/


static uint32_t rxcmd400( size_t len )
{
    size_t  i, j, shiftl, shiftr;

    __asm {
		setse1	#0b110<<6 | PIN_CLK    // trigger on high level
		waitse1    // for transition from driving CMD pin to reading CMD pin
//.loop0
//		testp	#PIN_CLK   wc
//	if_nc	jmp	#.loop0    // for transition from driving CMD pin to reading CMD pin

		dirl	#PIN_CMD
		wrpin	##P_SYNC_RX | (PIN_CLK - PIN_CMD & 7)<<24, #PIN_CMD    // setup responce smartpin mode
		wxpin	#15, #PIN_CMD    // CMD 16-bit shifter/buffer, pre-rising clock sampling

		wypin	#64, #PIN_CLK    // max of 64 clocks for response starting timeout, SD spec 4.12.4
		dirh	#PIN_CMD
		setse1	#0b001<<6 | PIN_CMD    // trigger on rising edge
		loc	ptrb, #resp    // response buffer address in hubRAM, static location

		mov	i, len    // early prep of response length
		add	i, #1    // round up
		shr	i, #1    // number of shortwords for response
		shl	i, #4    // number of clocks for response

		mov	j, #4    // 4 short words = max of 64 bits timeout for start-bit
.loop1
		waitse1
		rdpin	pb, #PIN_CMD
		rev	pb
		signx	pb, #15
		not	shiftr, pb   wz    // non-zero when start-bit found
	if_z	djnz	j, #.loop1    // keep looking for start-bit, up to 64 bits

	if_nz	wypin	i, #PIN_CLK    // response bit-clocks, SD spec 4.4 and 4.12.4
		encod	shiftr
		add	shiftr, #1
		mov	shiftl, shiftr
		subr	shiftl, #16

		shr	i, #4    // number of shortwords for response
.loop2
	if_nz	waitse1
	if_nz	rdpin	pa, #PIN_CMD
		rev	pa
		mov	j, pa
		shr	pa, shiftr
		shl	pb, shiftl
		or	pb, pa
		movbyts	pb, #0b00_01
	if_nz	wrword	pb, ptrb++
		mov	pb, j
	if_nz	djnz	i, #.loop2

		dirl	#PIN_CMD
    }

    if( i )
        return 0;    // timed-out

#ifdef _DIAG
    for( i = 0; i < len; i++ )
        printf( " %02x", resp[i] );
    printf( "   shl=%2d  shr=%2d", shiftl, shiftr );
#endif

    return 1;    // card has responded
}


static uint32_t sdcmd4( uint32_t cmd, uint32_t arg )
{
    uint32_t  crc;

#ifdef _DIAG
    printf( "CMD%d  %08x  > ", cmd & 0xff, arg );    // diag
#endif
    _wxpin( PIN_CLK, 4 | 2<<16 );
    _wrpin( PIN_CMD, P_SYNC_TX | P_OE | (PIN_CLK - PIN_CMD & 7)<<24 | P_INVERT_B );    // setup CMD output smartpin mode
    _wxpin( PIN_CMD, 31 );    // 32-bit shifter, continuous mode
    _wypin( PIN_CMD, _rev( (0x40 | cmd)<<24 | arg>>8 ) );    // first 32 bits into tx shifter, continuous mode

    _wypin( PIN_CLK, 6 * 8 );    // begin SD clocks, tx smartpin won't see clocks for about 8 sysclock ticks
    _pinl( PIN_CMD );    // start tx shifter, in continuous mode

    __asm {    // SD spec 4.5
		mov	pa, cmd
		or	pa, #0x40
		shl	pa, #24
		setq	pa
		mov	crc, #0
		crcnib	crc, #0x48    // CCITT polynomial is 1 + x3 + x7 (0x09 reversed for CRCNIB)
		crcnib	crc, #0x48
		setq	arg
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		rev	crc
		shr	crc, #24
		or	crc, #1    // stop-bit at end of command packet
    }
    _wypin( PIN_CMD, _rev( arg<<24 | crc<<16 ) );    // final 16 bits into tx buffer

    crc = rxcmd400( 6 );

    switch( cmd )  {
        case 17:
#ifdef _DIAG
            if( crc )
                crc = crc7( resp, 5 );
#endif
            crc = rxblock( buff );
            break;
    }

    if( crc )
        return 1;
#ifdef _DIAG
    puts( " TIMED-OUT!" );
#endif
    return 0;
}


static uint32_t sdcmd400( uint32_t cmd, uint32_t arg )
{
    uint32_t  crc;

    if( cmd & 0x100 )  {    // check if it's an ACMD
        if( !sdcmd400( 55, rca ) )    // ACMD's prefix
            return 0;
    }
#ifdef _DIAG
    printf( "CMD%d  %08x  > ", cmd & 0xff, arg );    // diag
#endif

    _wrpin( PIN_CMD, P_SYNC_TX | P_OE | (PIN_CLK - PIN_CMD & 7)<<24 );    // setup CMD output smartpin mode
    _wxpin( PIN_CMD, 31 );    // 32-bit shifter, continuous mode
    _wypin( PIN_CMD, _rev( (0x40 | cmd)<<24 | arg>>8 ) );    // first 32 bits into tx shifter, continuous mode
    _wypin( PIN_CLK, 6 * 8 );    // begin SD clocks, tx smartpin won't see clocks for about 8 sysclock ticks
    _pinl( PIN_CMD );    // start tx shifter, in continuous mode

    __asm {    // SD spec 4.5
		mov	pa, cmd
		or	pa, #0x40
		shl	pa, #24
		setq	pa
		mov	crc, #0
		crcnib	crc, #0x48    // CCITT polynomial is 1 + x3 + x7 (0x09 reversed for CRCNIB)
		crcnib	crc, #0x48
		setq	arg
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		rev	crc
		shr	crc, #24
		or	crc, #1    // stop-bit at end of command packet
    }
    _wypin( PIN_CMD, _rev( arg<<24 | crc<<16 ) );    // final 16 bits into tx buffer

    switch( cmd )  {
        case 2:
        case 9:
        case 10:
            crc = rxcmd400( 17 );
#ifdef _DIAG
            if( crc )
                crc = crc7( resp+1, 15 );
#endif
            break;
        case 3:
            crc = rxcmd400( 6 );
            if( crc )  {
#ifdef _DIAG
                crc = crc7( resp, 5 );
#endif
                rca = resp[1]<<24 | resp[2]<<16;
            }
            break;
        default:
            crc = rxcmd400( 6 );
#ifdef _DIAG
            if( crc )
                crc = crc7( resp, 5 );
#endif
            break;
    }
    if( crc )  {
#ifdef _DIAG
        printf( "   CRC = %02x\n", crc );
#endif
        return( resp[1]<<24 | resp[2]<<16 | resp[3]<<8 | resp[4] );
    }
#ifdef _DIAG
    puts( " TIMED-OUT!" );
#endif
    return 0;
}


static void  sdrelease( void )    // Targetted at controlling Roger Loh's 4-bit SD add-on board
{
    _waitms( 1 );
    _pinf( SD_BASE_PIN | 7<<6 );    // All eight I/O pins of the add-on header are used by the 4-bit SD board
    _wrpin( SD_BASE_PIN | 7<<6, 0 );
    _waitms( 1 );
}


static int  sdpower( void )
{
    sdrelease();

    if( !_pinr( PIN_CD ) )  {    // Card Detect will pull this down against the 22k ohm pull-up
        _pinh( PIN_CD );    // power off the SD card by overpowering the 1800 ohm Card Detect
        puts( " Card detected ... 40 ms power cycle of SD card ... " );
        _waitms( 40 );     // aproximately a 20 ms natural discharge curve for a plugged in card, SD spec 6.4.1
        _pinf( PIN_CD );     // and back on, released back to Card Detect function
        puts( " DONE!" );
        _waitms( 2 );
        return 1;
    }
    return 0;
}


static int  sdinit( void )    // Targetted at controlling Roger Loh's 4-bit SD add-on board
{
    uint32_t  rc;

    if( sdpower() )
    {
        _pinstart( PIN_LED, P_REG_UP | P_OE | P_INVERT_OUTPUT |    // outputs to Activity LED, doubles up as rx start detection
                (PIN_CLK - PIN_LED & 7)<<28 |    // samrtA input select, count clock edges
                (PIN_CMD - PIN_LED & 7)<<24,     // samrtB input select, count while CMD high
                0, 0 );

        rc = (_clockfreq() + 200_000) / 400_000;    // 400 kHz SD clock frequency for card init
        _pinstart( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT,    // SD clock gen smartpin
                rc | rc/2<<16, 0 );
        _waitms( 1 );
        _wypin( PIN_CLK, 80 );    // SD spec 6.4.1
        _waitms( 1 );
        printf( " Card Init 400 kHz Clock Divider = %d\n", rc );

        rca = 0;
        sdcmd400( 0, 0 );    // abort to "idle" state - in case the power cycle reset wasn't successful

        rc = sdcmd400( 8, 0x15a );    // SD spec 4.3.13
        if( rc == 0x15a )    // SD v2.0 valid voltage and check pattern
        {
            uint32_t  time, timer1 = _getms();
            uint32_t  timer2 = timer1;

            printf( "  R7 %08x - v2.0+ SD Card\n", rc );
            while(1)  {
                rc = sdcmd400( 0x100|41, 1<<30 | 1<<20 );   // indicate HCS capable and 3v3 voltage
                if( rc>>31 )    // valid Ready bit, card has switched from "idle" to "ready" state
                    break;

                do  {
                    time = _getms();
                } while( time - timer2 < 40 );    // pause for 40 ms, SD spec 4.4
                timer2 += 40;

                if( time - timer1 > 1000 )    // SD spec 4.2.3
                    break;
            }
            printf( "  OCR %08x\n", rc );

            // CMD11 voltage switch here for controllers and cards that support it
            //   - Prop2 best suited to default 3V3 signalling, don't add CMD11

            // Next two commands are explicitly ordered and required - singuluar operational path
            // If the sequence is not succesful then CMD3 will timeout and return a zero value
            sdcmd400( 2, 0 );   // switch from "ready" to "ident" state, also next card status indicates a partial ACMD state
            if( rc = sdcmd400( 3, 0 ) )    // switch from "ident" to "standby" state, also moves into "data transfer mode"
            {
                rca = rc & 0xffff_0000;
                printf( "  RCA %08x\n", rca );

                sdcmd400( 7, rca );    // SELECT - switch from "standby" to "tran" state
                if( sdcmd400( 0x100|6, 2 ) )  {    // Engage 4-bit parallel data mode
                    puts( " Card init complete" );
                    return  rca;
                }
            }
        }
    }
    puts( "Card not valid" );
    return  0;
}


//===================================================================================

void  main( void )
{
    uint32_t  rc;

    printf( "   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );
    _pinl( 56 );    // diag
/*
    _pinstart( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT,    // SD clock gen smartpin
                4 | 2<<16, 0 );
    __asm {
		wypin	#8, #PIN_CLK
		waitx	##50000
		setse1	#0b110<<6 | PIN_CLK    // trigger on high level
		waitse1    // for transition from driving CMD pin to reading CMD pin
    }
*/
    rc = sdinit();  // returns a valid RCA with card in data transfer "standby" state, or zero for a fail

    if( rc )
    {
        for( rc = 0; rc < 2000; rc++ )
            buff[rc] = 0;

        _pinh( 56 );    // diag
        rc = sdcmd4( 17, 0 );
        rc = sdcmd4( 17, 0 );

        puts( "All finished  :)" );
    }

    sdpower();
    _clkset( 1, 20_000 );  // cool running
}
